/**
 * Copyright (c) 2011 Colony Project
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

/**
 * @file cliffsensor.cpp
 * @brief Contains code to control the cliffsensor.
 *
 * Implementation of functions for cliffsensor use.
 *
 * @author Colony Project, CMU Robotics Club
 * @author Priyanka Deo
 **/

#include "ros/ros.h"
#include "cliffsensor.h"
#include <cstdlib>

/**
 * @defgroup cliffsensor Cliffsensor
 * @brief Functions for using the cliffsensors
 *
 * @{
 **/

/* Cliffsensor state variables
 */
static int front_raw; /**< The current raw value data of the front cliffsensor. */
static int left_raw; /**< The current raw value data of the left cliffsensor. */
static int back_raw; /**< The current raw value data of the right cliffsensor. */
static bool is_cliff; /**< Boolean to represent whether or not there is a cliff. */

/**
 * @brief Changes cliff sensor status
 *
 * Changes cliff sensor status based on subscription to topic cliff_status_changed
 *
 * @param msg The message from the cliff_status_changed topic, containing 
 * status of all cliff sensors.
 */
void changed_cliff_status(const cliffsensor::cliff_status_changed::ConstPtr& msg)
{
  return;
}

/**
 * @brief Outputs status of cliffsensors
 *
 * Serves the service query_cliff by responding to service requests with the
 * status of the cliff sensors.
 * @param req The request. The only field is the units requested.
 * @param res The response. The fields will be filled with values.
 */
bool cliff_query(cliffsensor::query_cliff::Request &req,
                  cliffsensor::query_cliff::Response &res)
{
    int threshold = 500; //TODO fix after figuring out what sensors give us
    res.front_raw = front_raw; 
    res.left_raw = left_raw;
    res.right_raw = back_raw;
    res.cliff_status = is_cliff;

    ROS_DEBUG("Cliffsensor status queried");
    return true;
}

/**
 * @brief Cliffsensor driver. This is a ROS node that controls cliffsensor status.
 *
 * This is the main function for the cliffsensors node. It is run when the node
 * starts and initializes the cliffsensors. It then subscribes to the
 * cliff_status_changed topics, and advertises the
 * query_cliff service.
 * 
 * @param argc The number of command line arguments (should be 1)
 * @param argv The array of command line arguments
 **/
int main(int argc, char **argv)
{
    /* Initialize in ROS the cliffsensor driver node */
    ros::init(argc, argv, "cliffsensor_driver");

    /* Advertise that this serves the query_cliff service */
    ros::NodeHandle n;
    ros::ServiceServer service = n.advertiseService("query_cliff",
                                                    cliff_query);

    /* Subscribe to the cliff_status_changed topic */
    ros::Subscriber sub0 = n.subscribe("cliff_status_changed", 
                                        4, changed_cliff_status);

    /* Initialize hardware for cliffsensors */
    // Hardware init functions here

    ROS_INFO("Ready to query cliffsensors.");
    ros::spin();

    return 0;
}

/** @} **/
